#include "prjlibs-c/standards.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <limits.h>
#include <errno.h>
#include <signal.h>
#include <dirent.h>
#include <stddef.h>
#include <time.h>

#include <skalibs/stddjb.h>
#include "prjlibs-c/constants.h"
#include "prjlibs-c/ascii.h"
#include "prjlibs-c/types.h"
#include "prjlibs-c/nsec.h"
#include "die.h"
#include "drop_root.h"
#include "fmt0.h"
#include "get_addr.h"
#include "must.h"
#include "read_line.h"
#include "respond.h"
#include "timeout.h"
#include "types.h"
#include "keys.h"

ccp PROG="posterity-imap";

static char const capab[]="CAPABILITY IMAP4rev1 AUTH=ANONYMOUS";
/* An IMAP4rev1 server is NOT required to support obsolete IMAP4 or IMAP2bis 
   commands, though some do */

typedef struct message message_t;
struct message {
  message_t* next;
  imap_uid_t uid;
  len_t name_len;
  flag_t flag_Seen:1;
  flag_t flag_Answered:1;
  flag_t flag_Flagged:1;
  flag_t flag_Deleted:1;
  flag_t flag_Draft:1;
  flag_t flag_Recent:1;
  flag_t selected:1;
  char name[1];
};
#define MESSAGE_SIZE(len) (offsetof(message_t, name)+(len)+1)
static message_t** msg_list=null;
static imap_uid_t n_msgs=0;

typedef struct mailbox mailbox_t;
struct mailbox {
  mailbox_t* next;
  flag_t flag_subscribed:1;
  flag_t flag_exists:1;
  len_t name_len;
  char name[1];
};
#define MAILBOX_SIZE(len) (offsetof(mailbox_t, name)+(len)+1)
static mailbox_t* mb_list=null, * mb_selected;

static char buffer_space[BUFFER_INSIZE];
static uint64_buf_t pid_buf;
static ccp remote_addr="", tag;
static stralloc cmd=GEN_ALLOC_ZERO, args=GEN_ALLOC_ZERO,
  path=GEN_ALLOC_ZERO, scratch=GEN_ALLOC_ZERO;
static flag_t flag_uid;
static struct taia mb_list_updated=TAIA_ZERO;

#define RESPOND_OK_0()       respond(tag," OK ",cmd.s," completed",null,null)
#define RESPOND_OK_1(r0)     respond(tag," OK ", (r0),null,null,null)
#define RESPOND_NO_1(r0)     respond(tag," NO ", (r0),null,null,null)
#define RESPOND_BAD_1(r0)    respond(tag," BAD ",(r0),null,null,null)
#define RESPOND_BAD_SYNTAX() respond(tag," BAD bad syntax",null,null,null,null)
#define RESPOND_BYE_1(r0)    respond("* BYE ",(r0),null,null,null,null)
#define RESPOND_NOTAG_1(r0)             respond("* ", (r0),null,null,null,null)
#define RESPOND_NOTAG_2(r0,r1)          respond("* ", (r0),(r1),null,null,null)
#define RESPOND_NOTAG_3(r0,r1,r2)       respond("* ", (r0),(r1),(r2),null,null)
#define RESPOND_NOTAG_5(r0,r1,r2,r3,r4) respond("* ", (r0),(r1),(r2),(r3),(r4))
#define RESPOND_CONT_0() respond("+ send data", null, null, null, null, null)

static void update_mb_list(void) {
  DIR* dir;
  struct stat st;
  struct taia last_mod=TAIA_ZERO, now;
  mailbox_t* new_list=null;
  taia_now(&now);
  if (stat(".", &st)!=0)
    strerr_die4sys(errstat, PROG, err_unable, err_stat, "\".\": ");
  tai_unix(taia_secp(&last_mod), st.st_mtime);
  taia_nano(&last_mod)=mtime_nsec(&st);
  if (taia_less(&last_mod, &mb_list_updated)) return;
  dir=opendir(".");
  if (dir==null)
    strerr_die4sys(errstat, PROG, err_unable, err_open, "\".\": ");
  for (;;) {
    struct dirent const* de;
    mailbox_t* mb, ** mbp;
    len_t len;
    errno=0;
    de=readdir(dir);
    if (de==null) {
      if (errno==0) break;
      strerr_die3sys(errstat, PROG, err_unable, "read \".\": ");
    }
    if (de->d_name[0]=='.') continue;
    len=str_len(de->d_name);
    for (mbp=&mb_list; *mbp!=null; mbp=&(*mbp)->next)
      if ((*mbp)->name_len==len && byte_equal(de->d_name, len, (*mbp)->name))
        break;
    if (*mbp!=null) {
      mb=*mbp;
      *mbp=mb->next;
      mb->flag_exists=1;
      mb->next=new_list;
      new_list=mb;
    } else {
      len_t size;
      ++len;
      if (len==0) die_overflow();
      size=MAILBOX_SIZE(len);
      if (size<len) die_overflow();
      mb=(void*)alloc(size);
      if (mb==null) die_alloc();
      mb->flag_subscribed=0;
      mb->flag_exists=1;
      byte_copy(mb->name, len, de->d_name);
      mb->name_len=len-1;
      mb->next=new_list;
      new_list=mb;
    }
  }
  closedir(dir);
  while (mb_list!=null) {
    mailbox_t* const next=mb_list->next;
    if (!mb_list->flag_subscribed) alloc_free(mb_list);
    else {
      mb_list->next=new_list;
      new_list=mb_list;
      new_list->flag_exists=0;
    }
    mb_list=next;
  }
  mb_list=new_list;
  mb_list_updated=now;
  --tai_sec(taia_secp(&mb_list_updated));
}

static status_t get_astring(stralloc* astring, stralloc* args,
                            flag_t flag_wild_ok, flag_t flag_last) {
  len_t i=0;
  if (args->len==0) return -1;
  astring->len=0;
  switch (args->s[0]) {
    case '{': {
      len_t len;
      ++i;
      i+=scan_uint(args->s+i, &len);
      if (i==1 || i+1!=args->len || args->s[i]!='}') return -1;
      must_ready(astring, len);
      RESPOND_CONT_0();
      for (;;) {
        status_t const cr=read_line(astring);
        if (astring->len==len) break;
        if (cr) MUST_CATS(astring, "\r");
        MUST_CATS(astring, "\n");
        if (astring->len>len) return -1;
      }
      if (flag_last) args->len=0;
      else read_line(args);
      return 0;
    }
    case '"':
      for (;;) {
        ++i;
        if (i==args->len) return -1;
        if (args->s[i]=='"') break;
        if (args->s[i]=='\\') ++i;
        if (i==args->len) return -1;
        must_append(astring, args->s+i);
      }
      ++i;
      break;
    default:
      while (i!=args->len    && args->s[i]>=0x20 && args->s[i]<0x7f &&
             args->s[i]!='(' && args->s[i]!=')'  && args->s[i]!='{' &&
             args->s[i]!=' ' && args->s[i]!='\\' && args->s[i]!='"' &&
             (flag_wild_ok ||  (args->s[i]!='%'  && args->s[i]!='*')))
        ++i;
      if (i==0) return -1;
      must_copyb(astring, args->s, i);
      break;
  }
  if (flag_last) {
    if (i!=args->len) return -1;
  } else {
    if (i==args->len || i+1==args->len || args->s[i]!=' ') return -1;
    ++i;
  }
  args->len-=i;
  byte_copy(args->s, args->len, args->s+i);
  return 0;
}

static mailbox_t* find_mailbox(stralloc const* name) {
  mailbox_t* mb;
  for (mb=mb_list; mb!=null; mb=mb->next)
    if (mb->name_len==name->len && byte_diff(name->s, name->len, mb->name)==0)
      break;
  return mb;
}

static void alarm_handler(int sig) {
  (void)sig;
  if (ndelay_on(1)<0)
    strerr_die3sys(errstat, PROG, err_unable, "set descriptor flags: ");
  RESPOND_BYE_1("timeout");
  _exit(0);
}

enum state {
  state_unauthenticated=1,
  state_authenticated  =2,
  state_selected       =4,
  states_authsel=state_authenticated|state_selected,
  states_all    =state_unauthenticated|state_authenticated|state_selected,
};
static enum state state=state_unauthenticated;

typedef void command_function(void);

typedef struct {
  ccp name;
  enum state legal_states;
  flag_t uid_ok:1;
  command_function* func;
} const command_t;

#define ASSERT_NO_ARGS() \
  do if (args.len!=0) { \
    RESPOND_BAD_1("this command does not accept arguments"); \
    return; \
  } while (0)

static void cmdx_readonly(void) {
  RESPOND_NO_1("this is a read-only service");
}

static mailbox_t* do_mailbox_prep(flag_t flag_last_arg) {
  stralloc arg=GEN_ALLOC_ZERO;
  mailbox_t* mb;
  update_mb_list();
  if (get_astring(&arg, &args, 0, flag_last_arg)!=0) {
    RESPOND_BAD_SYNTAX();
    stralloc_free(&arg);
    return null;
  }
  mb=find_mailbox(&arg);
  stralloc_free(&arg);
  if (mb==null) RESPOND_NO_1("no such mailbox");
  return mb;
}

static void do_list(flag_t flag_subscribed_only) {
  stralloc arg=GEN_ALLOC_ZERO;
  mailbox_t const* mb;
  flag_t flag_ref_empty;
  update_mb_list();
  if (get_astring(&arg, &args, 0, 0)!=0) goto bad;
  flag_ref_empty=(arg.len==0);
  if (get_astring(&arg,  &args, 1, 1)!=0) bad: {
    RESPOND_BAD_SYNTAX();
    stralloc_free(&arg);
    return;
  }
  if (arg.len==0)
    RESPOND_NOTAG_2(cmd.s, " (\\Noselect) NIL \"\"");
  else if (flag_ref_empty && byte_chr(arg.s, arg.len, '\0')==arg.len) {
    len_t i=byte_chr(arg.s, arg.len, '*');
    if (i==arg.len) i=byte_chr(arg.s, arg.len, '%');
    if (i==arg.len) {
      mb=find_mailbox(&arg);
      if (mb!=null)
        RESPOND_NOTAG_5(cmd.s, " (\\Noinferiors",
                        (mb->flag_exists? "": " \\Noselect"),
                        ") NIL ", mb->name);
      /* modified UTF-7 & */ 
    } else {
      /* multiple wildcards */ 
      len_t j=arg.len-i-1;
      for (mb=mb_list; mb!=null; mb=mb->next) {
        if (flag_subscribed_only && !mb->flag_subscribed) continue;
        if (mb->name_len<arg.len-1) continue;
        if (byte_diff(mb->name, i, arg.s)) continue;
        if (byte_diff(mb->name+mb->name_len-j, j, arg.s+i+1)) continue;
        RESPOND_NOTAG_5(cmd.s, " (\\Noinferiors",
                        (mb->flag_exists? "": " \\Noselect"),
                        ") NIL ", mb->name);
      }
    }
  }
  RESPOND_OK_0();
  stralloc_free(&arg);
}

static void do_subscribe(flag_t flag_subscribe) {
  mailbox_t* mb=do_mailbox_prep(1);
  if (mb==null) return;
  mb->flag_subscribed=flag_subscribe;
  RESPOND_OK_0();
}

static void writepart(char* sep, ccp s0, ccp s1) {
  mustwrite(sep);
  mustwrite(s0);
  mustwrite(s1);
  sep[0]=' ';
  sep[1]='\0';
}

static void cmd_capability(void) {
  ASSERT_NO_ARGS();
  RESPOND_NOTAG_1(capab);
  RESPOND_OK_0();
}

static void cmd_noop(void) {
  ASSERT_NO_ARGS();
  RESPOND_OK_0();
}

static void cmd_logout(void) {
  ASSERT_NO_ARGS();
  RESPOND_BYE_1("logging out");
  RESPOND_OK_0();
  _exit(0);
}

static char const auth_anon[]="anonymous";

static void cmd_authenticate(void) {
  if ((args.len!=sizeof auth_anon-1 ||
       case_diffb(args.s, args.len, auth_anon))) {
    RESPOND_NO_1("unknown authentication method");
    return;
  }
  RESPOND_CONT_0();
  read_line(&scratch); /*log*/ 
  RESPOND_OK_0();
  state=state_authenticated;
}

static void cmd_login(void) {
  flag_t flag_anon;
  if (get_astring(&scratch, &args, 0, 0)!=0) {
    RESPOND_BAD_SYNTAX();
    return;
  }
  flag_anon=(scratch.len==(sizeof auth_anon)-1 &&
             byte_equal(scratch.s, scratch.len, auth_anon));
  if (get_astring(&scratch, &args, 0, 1)!=0) {
    RESPOND_BAD_SYNTAX();
    return;
  }
  if (!flag_anon) RESPOND_NO_1("you must log in as \"anonymous\"");
  else {
    RESPOND_OK_0();
    state=state_authenticated;
  }
}

static void cmd_list(void) { do_list(0); }
static void cmd_lsub(void) { do_list(1); }
static void cmd_subscribe  (void) { do_subscribe(1); }
static void cmd_unsubscribe(void) { do_subscribe(0); }

static stralloc next_uid_sa=GEN_ALLOC_ZERO;

static imap_uid_t next_uid(ccp mb_name, len_t mb_len) {
  static char const uid_space[sizeof (uid_buf_t)]="";
  len_t uid_offset;
  imap_uid_t uid_opened, uid_linked;
  ino_t ino_opened;
  must_copyb(&next_uid_sa, mb_name, mb_len);
  MUST_CATS(&next_uid_sa, "/.posterity/uidlock/");
  uid_offset=next_uid_sa.len;
  MUST_CATS0(&next_uid_sa, uid_space);
  for (uid_opened=1;;) {
    struct stat st;
    imap_uid_t nlink;
    fmt_uid(next_uid_sa.s+uid_offset, uid_opened);
    if (stat(next_uid_sa.s, &st)!=0) {
      if (errno==ENOENT) return uid_opened;
      strerr_warn5sys(PROG, err_unable, err_stat, next_uid_sa.s, ": ");
      return 0;
    }
    ino_opened=st.st_ino;
    nlink=st.st_nlink;
    uid_linked=uid_opened+nlink;
    if (IS_NEGATIVE(nlink_t, st.st_nlink) || (nlink_t)nlink!=st.st_nlink ||
        uid_linked<nlink)
      die_overflow();
    fmt_uid(next_uid_sa.s+uid_offset, uid_linked);
    if (stat(next_uid_sa.s, &st)!=0) {
      if (errno==ENOENT) return uid_linked;
      strerr_warn5sys(PROG, err_unable, err_stat, next_uid_sa.s, ": ");
      return 0;
    }
    if (st.st_ino!=ino_opened) uid_opened=uid_linked; /* this is a new file */
    /* otherwise, we lost a race; try again */
  }
}

static void cmd_select(void) {
  uid_buf_t buf_uid;
  DIR* dir;
  imap_uid_t i, minuid=-1, maxuid=0;
  len_t msg_id_offset;
  message_t* msgs=null;
  state=state_authenticated;
  for (i=0; i!=n_msgs; ++i) alloc_free(msg_list[i]);
  alloc_free(msg_list);
  msg_list=null;
  n_msgs=0;
  mb_selected=do_mailbox_prep(1);
  if (mb_selected==null) return;
  must_copyb(&path, mb_selected->name, mb_selected->name_len);
  MUST_CATS0(&path, "/new");
  dir=opendir(path.s);
  if (dir==null) { RESPOND_NO_1("unable to open mailbox"); return; }
  must_copyb(&scratch, mb_selected->name, mb_selected->name_len);
  MUST_CATS(&scratch, "/.posterity/msginfo/");
  msg_id_offset=scratch.len;
  MUST_CATS0(&scratch, "");
  for (;;) {
    struct cdb c;
    struct dirent* de;
    len_t len_uid, len_id, size_id;
    imap_uid_t uid;
    message_t* msg;
    fd_t fd;
    errno=0;
    de=readdir(dir);
    if (de==null) {
      if (errno==0) break;
      strerr_warn5sys(PROG, err_unable, err_read, path.s, ": ");
      while (msgs!=null) {
        msg=msgs;
        msgs=msg->next;
        alloc_free(msg);
      }
      RESPOND_NO_1("unable to read mailbox");
      return;
    }
    if (de->d_name[0]=='.') continue;
    len_id=str_len(de->d_name)+1;
    if (len_id==0) die_overflow();
    scratch.len=msg_id_offset;
    must_catb(&scratch, de->d_name, len_id);
    fd=open(scratch.s, O_RDONLY);
    if (fd<0) {
      strerr_warn5sys(PROG, err_unable, err_open, scratch.s, ": ");
      continue;
    }
    if (cdb_init(&c, fd)!=0) {
      strerr_warn5sys(PROG, err_unable, err_read, scratch.s, ": ");
      continue;
    }
    switch (cdb_find(&c, (char*)key_uid, sizeof key_uid-1)) { 
      case 0:
        strerr_warn3x(PROG, "corrupt cdb: ", scratch.s);
        cdb_free(&c);
        fd_close(fd);
        continue;
      case -1:
        strerr_warn5sys(PROG, err_unable, err_read, scratch.s, ": ");
        cdb_free(&c);
        fd_close(fd);
        continue;
    }
    len_uid=cdb_datalen(&c);
    if (len_uid>=sizeof buf_uid) {
      strerr_warn3x(PROG, "corrupt cdb: ", scratch.s);
      cdb_free(&c);
      fd_close(fd);
      continue;
    }
    buf_uid[len_uid]='\0';
    if (cdb_read(&c, buf_uid, len_uid, cdb_datapos(&c))!=0) {
      strerr_warn5sys(PROG, err_unable, err_read, scratch.s, ": ");
      cdb_free(&c);
      fd_close(fd);
      continue;
    }
    cdb_free(&c);
    fd_close(fd);
    if (scan_uid(buf_uid, &uid)!=len_uid) {
      strerr_warn3x(PROG, "corrupt cdb: ", scratch.s);
      continue;
    }
    size_id=MESSAGE_SIZE(len_id);
    if (size_id<len_id) die_overflow();
    msg=(void*)alloc(size_id);
    if (msg==null) die_alloc();
    byte_copy(msg->name, len_id, de->d_name);
    msg->name_len=len_id-1;
    msg->flag_Seen=1;
    msg->flag_Answered=0;
    msg->flag_Flagged=0;
    msg->flag_Deleted=0;
    msg->flag_Draft=0;
    msg->flag_Recent=0;
    msg->uid=uid;
    msg->next=msgs;
    msgs=msg;
    ++n_msgs;
    if (minuid>uid) minuid=uid;
    if (maxuid<uid) maxuid=uid;
  }
  closedir(dir);
  if (n_msgs!=0) {
    message_t* msg;
    msg_list=(void*)alloc(n_msgs*sizeof *msg_list);
    if (msg_list==null) die_alloc();
    if (maxuid-minuid==n_msgs-1) /* no gaps; easy case */
      for (msg=msgs; msg!=null; msg=msg->next)
        msg_list[msg->uid-minuid]=msg;
    else {
      for (i=0; i!=n_msgs; ++i) msg_list[i]=null;
      for (msg=msgs; msg!=null; msg=msg->next) {
        i=(uint64)(msg->uid-minuid)*(n_msgs-1)/(maxuid-minuid);
        if (msg_list[i]==null) msg_list[i]=msg;
        else {
          imap_uid_t j=i;
          for (;;) {
            if (j!=n_msgs-1 && msg_list[++j]==null) {
              do {
                imap_uid_t const k=j-1;
                if (msg_list[k]->uid<msg->uid) break;
                msg_list[j]=msg_list[k];
                j=k;
              } while (j!=0);
              msg_list[j]=msg;
              break;
            } else if (i!=0 && msg_list[--i]==null) {
              do {
                imap_uid_t const k=i+1;
                if (msg_list[k]->uid>msg->uid) break;
                msg_list[i]=msg_list[k];
                i=k;
              } while (i!=n_msgs-1);
              msg_list[i]=msg;
              break;
            }
          }
        }
      }
    }
  }
  fmt0_uid(&buf_uid, n_msgs);
  RESPOND_NOTAG_2(buf_uid, " EXISTS");
  RESPOND_NOTAG_1("0 RECENT");
  RESPOND_NOTAG_1("FLAGS (\\Seen \\Answered \\Flagged \\Deleted \\Draft)");
  for (i=0; i!=n_msgs; ++i)
    if (!msg_list[i]->flag_Seen) {
      fmt0_uid(&buf_uid, i+1);
      RESPOND_NOTAG_3("OK [UNSEEN ", buf_uid, "] ok");
      break;
    }
  {
    imap_uid_t next=next_uid(mb_selected->name, mb_selected->name_len);
    if (next!=0) {
      fmt0_uid(&buf_uid, next);
      RESPOND_NOTAG_3("OK [UIDNEXT ", buf_uid, "] ok");
    }
  }
  RESPOND_NOTAG_1("OK [UIDVALIDITY 1] ok");
  RESPOND_NOTAG_1("OK [PERMANENTFLAGS ()] ok");
  RESPOND_OK_1("[READ_ONLY]");
  state=state_selected;
}

static void cmd_status(void) {
  uid_buf_t buf_uid;
  flag_t flag_recent=0, flag_unseen=0, flag_uidvalidity=0, flag_uidnext=0,
    flag_messages=0;
  len_t i;
  char sep[]=" (";
  mailbox_t const* mb=do_mailbox_prep(0);
  if (mb==null) return;
  if (args.s[0]!='(' || args.s[args.len-1]!=')') {
    RESPOND_BAD_SYNTAX();
    return;
  }
  args.s[args.len-1]=' ';
  for (i=1; i!=args.len;) {
    len_t j=byte_chr(args.s+i, args.len-i, ' ');
#define CASE(x) \
    if (j==sizeof #x-1 && !case_diffb(args.s+i, j, #x)) flag_##x=1; else
    CASE(recent)
    CASE(unseen)
    CASE(uidnext)
    CASE(uidvalidity)
    CASE(messages)
#undef CASE
      { RESPOND_BAD_SYNTAX(); return; }
    i+=j+1;
  }
  mustwrite("* STATUS ");
  mustwrite(mb->name);
  if (flag_recent)      writepart(sep, "RECENT 0", "");
  if (flag_uidvalidity) writepart(sep, "UIDVALIDITY 1", ""); 
  if (flag_unseen) {
    imap_uid_t msgi, n_unseen=0;
    for (msgi=0; msgi!=n_msgs; ++msgi)
      if (!msg_list[msgi]->flag_Seen) ++n_unseen;
    fmt0_uid(&buf_uid, n_unseen);
    writepart(sep, "UNSEEN ", buf_uid);
  }
  if (flag_uidnext) {
    imap_uid_t next=next_uid(mb->name, mb->name_len);
    if (next!=0) {
      fmt0_uid(&buf_uid, next);
      writepart(sep, "UIDNEXT ", buf_uid);
    }
  }
  if (flag_messages) {
    DIR* dir;
    imap_uid_t n_messages=0;
    len_t msg_id_offset;
    must_copyb(&scratch, mb->name, mb->name_len);
    MUST_CATS(&scratch, "/.posterity/msginfo/"); 
    msg_id_offset=scratch.len;
    must_copyb(&path, mb->name, mb->name_len);
    MUST_CATS0(&path, "/new");
    dir=opendir(path.s);
    if (dir==null) {
      strerr_warn5sys(PROG, err_unable, err_open, path.s, ": ");
      RESPOND_NO_1("unable to open mailbox");
      return;
    }
    for (;;) {
      struct dirent* de;
      struct stat st;
      errno=0;
      de=readdir(dir);
      if (de==null) {
        if (errno==0) break;
        strerr_warn5sys(PROG, err_unable, err_read, path.s, ": ");
        closedir(dir);
        RESPOND_NO_1("unable to read mailbox");
        return;
      }
      if (de->d_name[0]=='.') continue;
      scratch.len=msg_id_offset;
      must_cats0(&scratch, de->d_name);
      if (stat(scratch.s, &st)==0) ++n_messages;
      else if (errno!=ENOENT)
        strerr_warn5sys(PROG, err_unable, err_open, scratch.s, ": ");
    }
    closedir(dir);
    fmt0_uid(&buf_uid, n_messages);
    writepart(sep, "MESSAGES ", buf_uid);
  }
  respond(")", null, null, null, null, null);
  RESPOND_OK_0();
}

static void cmd_check(void) {
  ASSERT_NO_ARGS();
  /* sync memory with fs */ 
  RESPOND_OK_0();
}

static void cmd_close(void) {
  imap_uid_t i;
  ASSERT_NO_ARGS();
  state=state_authenticated;
  mb_selected=null;
  for (i=0; i!=n_msgs; ++i) alloc_free(msg_list[i]);
  alloc_free(msg_list);
  msg_list=null;
  n_msgs=0;
  RESPOND_OK_0();
}

/*keep open fd & sought offset for partial fetches*/ 
static void cmd_fetch(void) {
  flag_t flag_need_stat=0, flag_need_buffer=0, flag_need_cdb=0,
    flag_do_envelope=0, flag_do_flags=0, flag_do_body=0,
    flag_do_bodystructure=0, flag_do_internaldate=0, flag_do_rfc822=0,
    flag_do_rfc822_header=0, flag_do_rfc822_size=0, flag_do_rfc822_text=0,
    flag_do_uid=flag_uid;
  imap_uid_t i;
  if (args.len==0) { RESPOND_BAD_SYNTAX(); return; }
  if (n_msgs==0) { RESPOND_OK_0(); return; }
  MUST_CATST(&args, "");
  for (i=0; i!=n_msgs; ++i) msg_list[i]->selected=0;
  for (i=0;;) {
#define SCAN(s,n) \
   (*(s)=='*'? (*(n)=(flag_uid? msg_list[n_msgs-1]->uid: n_msgs), 1): \
               scan_uint((s), (n)))
    imap_uid_t msgi, msgj;
    len_t len=SCAN(args.s+i, &msgi);
    if (len==0) { RESPOND_BAD_SYNTAX(); return; }
    i+=len;
    if (args.s[i]!=':') msgj=msgi;
    else {
      ++i;
      len=SCAN(args.s+i, &msgj);
      if (len==0) { RESPOND_BAD_SYNTAX(); return; }
      i+=len;
    }
#undef SCAN
    if (msgj<msgi) {
      imap_uid_t tmp=msgj;
      msgj=msgi;
      msgi=tmp;
    }
    if (!flag_uid) {
      if (msgi!=0) --msgi;
      if (msgj!=0) msgj=msg_list[(msgj<n_msgs? msgj: n_msgs)-1]->uid;
    }
    else if (msgi<msg_list[0       ]->uid) msgi=0;
    else if (msgi>msg_list[n_msgs-1]->uid) msgi=n_msgs;
    else {
      imap_uid_t start=0, end=n_msgs-1, idx=0,
        startuid=msg_list[start]->uid, enduid=msg_list[end]->uid;
      while (start!=end) {
        message_t const* msg;
        idx=start+(uint64)(msgi-startuid)*(end-start)/(enduid-startuid);
        msg=msg_list[idx];
        if (msgi==msg->uid) break;
        if (msgi>msg->uid) {
          if (idx==end) break;
          start=idx+1;
          startuid=msg_list[start]->uid;;
          if (startuid>msgi) { idx=start; break; }
        } else {
          if (idx==start) break;
          end=idx-1;
          enduid=msg_list[end]->uid;
          if (enduid<msgi) break;
        }
      }
      if (msgi<=msg_list[idx]->uid) msgi=idx;
      else msgi=n_msgs;
    }
    while (msgi<n_msgs) {
      message_t* msg=msg_list[msgi];
      if (msg->uid>msgj) break;
      msg->selected=1;
      ++msgi;
    }
    if (args.s[i]!=',') break;
    ++i;
  }
  if (args.s[i]!=' ') { RESPOND_BAD_SYNTAX(); return; }
  ++i;
  if (args.len==i) { RESPOND_BAD_SYNTAX(); return; }
#define CASE(x) \
  (args.len-i==sizeof #x-1 && case_equalb(args.s+i, sizeof #x-1, #x))
  if (CASE(all)) {
    flag_do_flags=
      flag_do_internaldate=
      flag_do_rfc822_size=
      flag_do_envelope=1;
  } else if (CASE(full)) {
    flag_do_flags=
      flag_do_internaldate=
      flag_do_rfc822_size=
      flag_do_envelope=
      flag_do_body=1;
  } else if (CASE(fast)) {
    flag_do_flags=
      flag_do_internaldate=
      flag_do_rfc822_size=1;
#undef CASE
  } else {
    if (args.s[i]!='(') MUST_CATS(&args, " ");
    else {
      ++i;
      if (args.len-i<2 || args.s[args.len-1]!=')') {
        RESPOND_BAD_SYNTAX();
        return;
      }
      args.s[args.len-1]=' ';
    }
    while (i!=args.len) {
      if (0) (void)0;
#define CASE1(attrname) CASE2(attrname,attrname)
#define CASE2(attrname,flagname) \
      else if (args.len-i>=sizeof #attrname && \
               case_equalb(args.s+i, sizeof #attrname, #attrname " ")) do { \
        flag_do_##flagname=1; \
        i+=sizeof #attrname; \
      } while(0)
      CASE1(envelope);
      CASE1(flags);
      CASE1(internaldate);
      CASE1(uid);
      CASE1(rfc822);
      CASE2(rfc822.header, rfc822_header);
      CASE2(rfc822.size,   rfc822_size);
      CASE2(rfc822.text,   rfc822_text);
      CASE1(body);
      CASE1(bodystructure);
#undef CASE1
#undef CASE2
      else {
        /*
body[]
body[header]
body[header.fields (to astring)]
body[header.fields.not (to)]
body[text]
body[1.2.3.4.mime]
body[1.2.3.4.header]
body[1.2.3.4.header.fields (to)]
body[1.2.3.4.header.fields.not (to)]
body[1.2.3.4.text]
body.peek[]
body.peek[header]
body.peek[header.fields (to)]
body.peek[header.fields.not (to)]
body.peek[text]
body.peek[1.2.3.4.mime]
body.peek[1.2.3.4.header]
body.peek[1.2.3.4.header.fields (to)]
body.peek[1.2.3.4.header.fields.not (to)]
body.peek[1.2.3.4.text]
body.peek[1.2.3.4.text]<0.1>
 */ 
        { RESPOND_BAD_SYNTAX(); return; }
      }
    }
  }
  if (flag_do_internaldate || flag_do_rfc822 || flag_do_rfc822_header ||
      flag_do_rfc822_size)
    flag_need_stat=1;
  if (flag_do_rfc822 || flag_do_rfc822_header) flag_need_buffer=1;
  if (flag_do_envelope) flag_need_cdb=1;
  for (i=0; i!=n_msgs; ++i) {
    struct stat st;
    buffer msgbuf;
    fd_t fd;
    message_t const* msg=msg_list[i];
    uid_buf_t buf_uid;
    char sep[]="(";
    flag_t flag_do_flags_once=flag_do_flags;
    if (!msg->selected) continue;
    mustwrite("* ");
    fmt0_uid(&buf_uid, i+1);
    mustwrite(buf_uid);
    mustwrite(" FETCH ");
    must_copyb(&path, mb_selected->name, mb_selected->name_len);
    MUST_CATS(&path, "/new/");
    must_catb(&path, msg->name, msg->name_len+1);
    if (flag_need_stat)
      if (stat(path.s, &st)!=0) {
        strerr_warn5sys(PROG, err_unable, err_stat, path.s, ": ");
        continue;
      }
    if (flag_need_buffer) {
      fd=open2(path.s, O_RDONLY);
      if (fd<0) {
        strerr_warn5sys(PROG, err_unable, err_open, path.s, ": ");
        continue;
      }
      buffer_init(&msgbuf, &buffer_unixread, fd, buffer_space,
                  sizeof buffer_space);
    }
    if (flag_need_cdb) { 
    }
    if (flag_do_uid) {
      fmt0_uid(&buf_uid, msg->uid);
      writepart(sep, "UID ", buf_uid);
    }
    if (flag_do_envelope) { 
/*
envelope        ::= "(" env_date SPACE env_subject SPACE env_from
                    SPACE env_sender SPACE env_reply_to SPACE env_to
                    SPACE env_cc SPACE env_bcc SPACE env_in_reply_to
                    SPACE env_message_id ")"
env_bcc         ::= "(" 1*address ")" / nil
env_cc          ::= "(" 1*address ")" / nil
env_date        ::= nstring
env_from        ::= "(" 1*address ")" / nil
env_in_reply_to ::= nstring
env_message_id  ::= nstring
env_reply_to    ::= "(" 1*address ")" / nil
env_sender      ::= "(" 1*address ")" / nil
env_subject     ::= nstring
env_to          ::= "(" 1*address ")" / nil
address         ::= "(" addr_name SPACE addr_adl SPACE addr_mailbox
                    SPACE addr_host ")"
addr_adl        ::= nstring
                    ;; Holds route from [RFC-822] route-addr if
                    ;; non-NIL
addr_host       ::= nstring
                    ;; NIL indicates [RFC-822] group syntax.
                    ;; Otherwise, holds [RFC-822] domain name
addr_mailbox    ::= nstring
                    ;; NIL indicates end of [RFC-822] group; if
                    ;; non-NIL and addr_host is NIL, holds
                    ;; [RFC-822] group name.
                    ;; Otherwise, holds [RFC-822] local-part
addr_name       ::= nstring
                    ;; Holds phrase from [RFC-822] mailbox if
                    ;; non-NIL
*/
    }
    if (flag_do_body) { 
    }
    if (flag_do_bodystructure) { 
    }
    if (flag_do_internaldate) {
      struct tm* tm;
      tm=gmtime(&st.st_mtime);
      if (tm==null) strerr_warn3x(PROG, "gmtime failed: ", path.s);
      else {
        uint64_buf_t buf_date;
        static char months[12][4]=
          { "Jan", "Feb", "Mar", "Apr", "May", "Jun",
            "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
        fmt0_uint64(&buf_date, tm->tm_mday);
        writepart(sep, "INTERNALDATE \"", buf_date);
        mustwrite("-");
        mustwrite(months[tm->tm_mon]);
        mustwrite("-");
        fmt0_uint64(&buf_date, tm->tm_year+1900);
        mustwrite(buf_date);
        mustwrite(" ");
        fmt0_uint64(&buf_date, tm->tm_hour);
        if (buf_date[1]=='\0') mustwrite("0");
        mustwrite(buf_date);
        mustwrite(":");
        fmt0_uint64(&buf_date, tm->tm_min);
        if (buf_date[1]=='\0') mustwrite("0");
        mustwrite(buf_date);
        mustwrite(":");
        fmt0_uint64(&buf_date, tm->tm_sec);
        if (buf_date[1]=='\0') mustwrite("0");
        mustwrite(buf_date);
        mustwrite(" +0000\"");
      }
    }
    if (flag_do_rfc822) {
      fsize_buf_t buf_fsize;
      fmt0_fsize(&buf_fsize, st.st_size);
      writepart(sep, "RFC822 {", buf_fsize);
      mustwrite("}\r\n");
      if (buffer_flush(buffer_1)<0) die_write("to network");
      for (;;) {
        len_t const len=must_fill(&msgbuf, path.s);
        if (len==0) break;
        if (allwrite(1, buffer_RPEEK(&msgbuf), len)<0) die_write("to network");
        buffer_RSEEK(&msgbuf, len);
      }
    }
    if (flag_do_rfc822_header) { 
/*
* x FETCH (BODY[HEADER] {46}
Content-Type: TEXT/PLAIN; CHARSET=US-ASCII

)
*/
    }
    if (flag_do_rfc822_size) {
      fsize_buf_t buf_fsize;
      fmt0_fsize(&buf_fsize, st.st_size);
      writepart(sep, "RFC822.SIZE ", buf_fsize);
    }
    if (flag_do_rfc822_text) { 
    }
    if (flag_do_flags_once) {
      char flagsep[]="(";
      writepart(sep, "FLAGS ", "");
#define CASE(f) if (msg->flag_##f) writepart(flagsep, "\\" #f, "")
      CASE(Seen);
      CASE(Answered);
      CASE(Flagged);
      CASE(Deleted);
      CASE(Draft);
      CASE(Recent);
#undef CASE
      if (flagsep[0]=='(') mustwrite(flagsep);
      mustwrite(")");
    }
    respond(")", null, null, null, null, null);
    if (flag_need_buffer)
      if (fd_close(fd)!=0)
        strerr_warn5sys(PROG, err_unable, err_close, path.s, ": ");
  }
  RESPOND_OK_0();
}

static void cmd_search(void) {
  static char const charset[]="charset ";
  if (args.len>=sizeof charset-1 &&
      case_equalb(args.s, sizeof charset-1, charset)) {
    RESPOND_NO_1("[BADCHARSET] charsets are not supported"); /* search */ 
    return;
  }
  /* arguments: [charset?, criteria*] */
  RESPOND_NO_1("command unimplemented"); /* search */ 
}

static command_t commands[]={
#define CMD1(states, uid_ok, name)        CMD2(states, uid_ok, name, name)
#define CMD2(states, uid_ok, name, alias) { #name,states,uid_ok,cmd_##alias }
#define CMDX(states, uid_ok, name)        { #name,states,uid_ok,cmdx_readonly }
  CMD1(states_all,            0, capability),
  CMD1(states_all,            0, noop),
  CMD1(states_all,            0, logout),
  CMD1(state_unauthenticated, 0, authenticate),
  CMD1(state_unauthenticated, 0, login),
  CMD1(states_authsel,        0, select),
  CMD2(states_authsel,        0, examine, select),
  CMDX(states_authsel,        0, create),
  CMDX(states_authsel,        0, delete),
  CMDX(states_authsel,        0, rename),
  CMD1(states_authsel,        0, subscribe),
  CMD1(states_authsel,        0, unsubscribe),
  CMD1(states_authsel,        0, list),
  CMD1(states_authsel,        0, lsub),
  CMD1(states_authsel,        0, status),
  CMDX(states_authsel,        0, append),
  CMD1(state_selected,        0, check),
  CMD1(state_selected,        0, close),
  CMDX(state_selected,        0, expunge),
  CMD1(state_selected,        1, search),
  CMD1(state_selected,        1, fetch),
  CMDX(state_selected,        1, store), 
  CMDX(state_selected,        1, copy),
#undef CMD1
#undef CMD2
#undef CMDX
};

static unsigned int const n_cmds=sizeof commands/sizeof commands[0];

int main(int argc, char** argv) {
  stralloc line=GEN_ALLOC_ZERO;
  struct sigaction action;

  (void)argv;
  if (argc>1)
    strerr_die3x(100, usage, "tcpserver 0 143 ", PROG);

  drop_root();
  fmt0_uint64(&pid_buf, getpid());
  remote_addr=get_addr();
  if (sigemptyset(&action.sa_mask)!=0)
    strerr_die5sys(100, PROG, err_unable, err_empty, err_sigset, ": ");
  action.sa_flags=0;
  action.sa_handler=alarm_handler;
  if (sigaction(SIGALRM, &action, null)!=0)
    strerr_die4sys(100, PROG, err_unable, err_sigact, ": ");
  RESPOND_NOTAG_3("OK [", capab, "] posterity");
  timeout_setup();

  for (;;) {
    len_t cmd_idx, args_idx;
    unsigned int i;
    line.len=0;
    read_line(&line);

    timeout_reset();

    if (line.len==0 || line.s[0]==' ') tag="*";
    else tag=line.s;
    cmd_idx=byte_chr(line.s, line.len, ' ');
    if (cmd_idx==line.len) {
      RESPOND_BAD_1("no command given");
      continue;
    }
    line.s[cmd_idx]='\0';
    ++cmd_idx;
    if (line.len-cmd_idx>=4 && !case_diffb(line.s+cmd_idx, 4, "uid ")) {
      flag_uid=1;
      cmd_idx+=4;
    } else
      flag_uid=0;
    args_idx=byte_chr(line.s+cmd_idx, line.len-cmd_idx, ' ');
    must_copyb(&cmd, line.s+cmd_idx, args_idx);
    MUST_CATS0(&cmd, "");
    args_idx+=cmd_idx;
    if (args_idx!=line.len) ++args_idx;
    must_copyb(&args, line.s+args_idx, line.len-args_idx);

    for (i=0; i!=n_cmds; ++i)
      if (case_equals(cmd.s, commands[i].name)) break;

    if (i==n_cmds)
      RESPOND_BAD_1("unknown command");
    else if ((commands[i].legal_states&state)!=state)
      RESPOND_BAD_1("that command is not valid in this state");
    else if (flag_uid && !commands[i].uid_ok)
      RESPOND_BAD_1("that command does not accept the UID modifier");
    else {
      command_function* func=commands[i].func;
      case_uppers(cmd.s);
      if (func!=cmd_select && func!=cmd_close && state==state_selected) {
        /* A server MUST send mailbox size updates automatically if a mailbox
           size change is observed during the processing of a command.
           Special rules exist for server notification of a client about the
           removal of messages to prevent synchronization errors; see the
           description of the EXPUNGE response for more detail. */
        /* EXISTS EXPUNGE(?) RECENT */ 
      }
      func();
      stralloc_free(&line);
      stralloc_free(&cmd);
      stralloc_free(&args);
      stralloc_free(&path);
      stralloc_free(&scratch);
    }
  }
}

/* Other compatibility issues with IMAP2bis, the most common variant of
   the earlier protocol, are discussed in [IMAP-COMPAT]. */
/*
       Characters are 7-bit US-ASCII unless otherwise specified.  Other
       character sets are indicated using a "CHARSET", as described in
       [MIME-IMT] and defined in [CHARSET].  CHARSETs have important
       additional semantics in addition to defining character set; refer
       to these documents for more detail.

   In modified UTF-7, printable US-ASCII characters except for "&"
   represent themselves; that is, characters with octet values 0x20-0x25
   and 0x27-0x7e.  The character "&" (0x26) is represented by the two-
   octet sequence "&-".

   All other characters (octet values 0x00-0x1f, 0x7f-0xff, and all
   Unicode 16-bit octets) are represented in modified BASE64, with a
   further modification from [UTF-7] that "," is used instead of "/".
   Modified BASE64 MUST NOT be used to represent any printing US-ASCII
   character which can represent itself.

   "&" is used to shift to modified BASE64 and "-" to shift back to US-
   ASCII.  All names start in US-ASCII, and MUST end in US-ASCII (that
   is, a name that ends with a Unicode 16-bit octet MUST end with a "-
   ").      For example, here is a mailbox name which mixes English, Japanese,
      and Chinese text: ~peter/mail/&ZeVnLIqe-/&U,BTFw-
*/
/*
4.1.2 The server MAY allow the EXPUNGE of a multi-accessed mailbox,
      and on subsequent FETCH commands return FETCH responses only for
      non-expunged messages and a tagged NO.

   After receiving a tagged NO FETCH response, the client SHOULD issue a
   NOOP command so that it will be informed of any pending EXPUNGE
   responses.  The client may then either reissue the failed FETCH
   command, or by examining the EXPUNGE response from the NOOP and the
   FETCH response from the FETCH, determine that the FETCH failed
   because of pending expunges.

   Example:  (Building upon the scenario outlined in 4.1.)

   <Client #2 attempts to FETCH a mix of expunged and non-expunged
   messages.  A FETCH response is returned only for then non-expunged
   messages along with a tagged NO>

             C2: B001 FETCH 3:5 ENVELOPE
             S2: * 3 FETCH ENVELOPE . . . (ENVELOPE info returned)
             S2: B001 NO Some of the requested messages no longer exist

   <Upon receiving a tagged NO FETCH response, Client #2 issues a NOOP
   to be informed of any pending EXPUNGE responses>

             C2: B002 NOOP
             S2: * 4 EXPUNGE
             S2: * 4 EXPUNGE
             S2: * 4 EXPUNGE
             S2: * 4 EXPUNGE
             S2: * 3 EXISTS
             S2: B002 OK NOOP Completed.

   <By receiving a FETCH response for message 3, and an EXPUNGE response
   that indicates messages 4:7 have been expunged, the client does not
   need to re-issue the FETCH>

4.2.1 If the ".SILENT" suffix is used, and the STORE completed
      successfully for all the non-expunged messages, the server SHOULD
      return a tagged OK.
4.2.2. If the ".SILENT" suffix is not used, and only expunged messages
       are referenced, the server SHOULD return only a tagged NO.
4.2.3. If the ".SILENT" suffix is not used, and a mixture of expunged
       and non-expunged messages are referenced, the server MAY set the
       flags and return a FETCH response for the non-expunged messages
       along with a tagged NO.
   A server MAY simply not return a search response for messages that
   have been expunged and it has not been able to inform the client
   about.
4.4.2 The server MAY allow the COPY of messages in a multi-access
      mailbox that contains expunged messages.

   Pending EXPUNGE response(s) MUST be returned to the COPY command.
*/
